from __future__ import absolute_import, division, print_function, unicode_literals

import sys
import logging
logger = logging.getLogger(__name__)

from ufl.equation import Equation
from ufl import Form
from dune.generator import Constructor, Method
import dune.common.checkconfiguration as checkconfiguration
import dune

def bbdgSpace(view, order=1, scalar=False, dimRange=None, field="double", storage="adaptive"):
    """create a discontinous galerkin space over an agglomerated grid

    Args:
        view: the underlying grid view
        order: polynomial order of the finite element functions
        dimRange: dimension of the range space
        field: field of the range space
        storage: underlying linear algebra backend

    Returns:
        Space: the constructed Space
    """

    from dune.fem.space import module, addStorage
    if dimRange is None:
        dimRange = 1
        scalar = True

    if dimRange > 1 and scalar:
        raise KeyError(\
                "trying to set up a scalar space with dimRange = " +\
                str(dimRange) + ">1")
    if dimRange < 1:
        raise KeyError(\
            "Parameter error in DiscontinuosGalerkinSpace with "+
            "dimRange=" + str(dimRange) + ": " +\
            "dimRange has to be greater or equal to 1")
    if order < 0:
        raise KeyError(\
            "Parameter error in DiscontinuousGalerkinSpace with "+
            "order=" + str(order) + ": " +\
            "order has to be greater or equal to 0")
    if field == "complex":
        field = "std::complex<double>"

    agglomerate = view.hierarchicalGrid.agglomerate

    includes = [ "dune/vem/agglomeration/dgspace.hh" ] + view._includes
    dimw = view.dimWorld
    viewType = view._typeName

    gridPartName = "Dune::FemPy::GridPart< " + view._typeName + " >"
    typeName = "Dune::Vem::AgglomerationDGSpace< " +\
      "Dune::Fem::FunctionSpace< double, " + field + ", " + str(dimw) + ", " + str(dimRange) + " >, " +\
      gridPartName + ", " + str(order) + " >"

    constructor = Constructor(
                   ['pybind11::object gridView',
                    'const pybind11::function agglomerate'],
                   ['auto agglo = new Dune::Vem::Agglomeration<' + gridPartName + '>',
                    '         (Dune::FemPy::gridPart<' + viewType + '>(gridView), [agglomerate](const auto& e) { return agglomerate(e).template cast<unsigned int>(); } ); ',
                    'auto obj = new DuneType( *agglo );',
                    'pybind11::cpp_function remove_agglo( [ agglo ] ( pybind11::handle weakref ) {',
                    '  delete agglo;',
                    '  weakref.dec_ref();',
                    '} );',
                    '// pybind11::handle nurse = pybind11::detail::get_object_handle( &obj, pybind11::detail::get_type_info( typeid( ' + typeName + ' ) ) );',
                    '// assert(nurse);',
                    'pybind11::weakref( agglomerate, remove_agglo ).release();',
                    'return obj;'],
                   ['"gridView"_a', '"agglomerate"_a',
                    'pybind11::keep_alive< 1, 2 >()'] )

    spc = module(field, includes, typeName, constructor, scalar=scalar, storage=storage,ctorArgs=[view, agglomerate])
    addStorage(spc, storage)
    return spc.as_ufl()

from dune.fem.scheme import dg,galerkin
def bbdgScheme(model, space=None, penalty=1, solver=None, parameters={}):
    if space == None:
        try:
            space = model.lhs.arguments()[0].ufl_function_space()
        except AttributeError:
            raise ValueError("no space provided and could not deduce from form provided")
    spaceType = space._typeName
    penaltyClass = "Dune::Vem::BBDGPenalty<"+spaceType+">"
    return dg(model,space,penalty,solver,parameters,penaltyClass)
    # return galerkin(model,space,solver,parameters)

def vemSpace(view, order=1, testSpaces=None, scalar=False,
             dimRange=None, conforming=True, field="double",
             storage="adaptive", basisChoice=2,
             edgeInterpolation=False):
    """create a virtual element space over an agglomerated grid

    Args:
        view: the underlying grid view
        order: polynomial order of the finite element functions
        dimRrange: dimension of the range space
        field: field of the range space
        storage: underlying linear algebra backend

    Returns:
        Space: the constructed Space
    """

    from dune.fem.space import module, addStorage
    if dimRange is None:
        dimRange = 1
        scalar = True

    if dimRange > 1 and scalar:
        raise KeyError(\
                "trying to set up a scalar space with dimRange = " +\
                str(dimRange) + ">1")
    if dimRange < 1:
        raise KeyError(\
            "Parameter error in DiscontinuosGalerkinSpace with "+
            "dimRange=" + str(dimRange) + ": " +\
            "dimRange has to be greater or equal to 1")
    if order < 0:
        raise KeyError(\
            "Parameter error in DiscontinuousGalerkinSpace with "+
            "order=" + str(order) + ": " +\
            "order has to be greater or equal to 0")
    if field == "complex":
        field = "std::complex<double>"

    if testSpaces is None:
        if conforming:
            testSpaces = [  [0], [order-2], [order-2] ]
        else:
            testSpaces = [ [-1], [order-1], [order-2] ]
    if testSpaces == "serendipity":
        etaMin = view.hierarchicalGrid.agglomerate.minEdgeNumber
        testSpaces = [  [0], [order-2], [max(-1,order-etaMin)] ]
        print("serendipity:",testSpaces,etaMin)

    for i,ts in enumerate(testSpaces):
        if type(ts) == int:
            testSpaces[i] = [ts]

    agglomerate = view.hierarchicalGrid.agglomerate

    includes = [ "dune/vem/space/agglomeration.hh" ] + view._includes
    dimw = view.dimWorld
    viewType = view._typeName

    gridPartName = "Dune::FemPy::GridPart< " + view._typeName + " >"
    typeName = "Dune::Vem::AgglomerationVEMSpace< " +\
      "Dune::Fem::FunctionSpace< double, " + field + ", " + str(dimw) + ", " + str(dimRange) + " >, " +\
      gridPartName + ", " + str(order) + " >"

    constructor = Constructor(
                   ['pybind11::object gridView',
                    'const pybind11::function agglomerate',
                    'const std::array<std::vector<int>,'+str(view.dimension+1)+'> &testSpaces',
                    'int basisChoice','bool edgeInterpolation'],
                   ['auto agglo = new Dune::Vem::Agglomeration<' + gridPartName + '>',
                    '         (Dune::FemPy::gridPart<' + viewType + '>(gridView), [agglomerate](const auto& e) { return agglomerate(e).template cast<unsigned int>(); } ); ',
                    'auto obj = new DuneType( *agglo, testSpaces, basisChoice, edgeInterpolation );',
                    'pybind11::cpp_function remove_agglo( [ agglo ] ( pybind11::handle weakref ) {',
                    '  delete agglo;',
                    '  weakref.dec_ref();',
                    '} );',
                    '// pybind11::handle nurse = pybind11::detail::get_object_handle( &obj, pybind11::detail::get_type_info( typeid( ' + typeName + ' ) ) );',
                    '// assert(nurse);',
                    'pybind11::weakref( agglomerate, remove_agglo ).release();',
                    'return obj;'],
                   ['"gridView"_a', '"agglomerate"_a', '"testSpaces"_a',
                    '"basisChoice"_a', '"edgeInterpolation"_a',
                    'pybind11::keep_alive< 1, 2 >()'] )
    diameterMethod = Method('diameters',
       '''[]( DuneType &self ) { return self.blockMapper().indexSet().diameters(); }''' )

    spc = module(field, includes, typeName, constructor, diameterMethod,
                scalar=scalar, storage=storage,
                ctorArgs=[view, agglomerate, testSpaces, basisChoice, edgeInterpolation])
    addStorage(spc, storage)
    return spc.as_ufl()

def space(view, order=1, dimRange=None,
          testSpaces=None,
          conforming=True,
          version=None,
          scalar=False,
          basisChoice=2,
          field="double", storage="adaptive"):
    '''
        version is tuple,list for a general vem space (=tests[aces)
        version == "dg" (bbdg)
        version == "dgSimplex" (dg on triangulation)
        version == "continuous" (vem)
        version == "continuousSimplex" (lagrange)
        version == "non-conformin" (vem)
    '''
    if version is None and testSpaces is None:
        raise AttributeError("version and testSpaces parameters can not both be None (default)")
    if version is None: version = testSpaces
    if isinstance(version,tuple) or isinstance(version,list):
        return vemSpace(view,order=order,scalar=scalar,dimRange=dimRange,field=field,storage=storage,
                        testSpaces=version, basisChoice=basisChoice)
    elif version == "dg":
        return bbdgSpace(view,order=order,scalar=scalar,dimRange=dimRange,field=field,storage=storage)
    elif version == "dgSimplex":
        from dune.fem.space import onb
        return onb(view,order=order,scalar=scalar,dimRange=dimRange,field=field,storage=storage)
    elif version == "continuous":
        return vemSpace(view,order=order,scalar=scalar,dimRange=dimRange,field=field,storage=storage,
                        conforming=True, basisChoice=basisChoice)
    elif version == "continuousSimplex":
        from dune.fem.space import lagrange
        return lagrange(view,order=order,scalar=scalar,dimRange=dimRange,field=field,storage=storage)
    elif version == "non-conforming":
        return vemSpace(view,order=order,scalar=scalar,dimRange=dimRange,field=field,storage=storage,
                        conforming=False, basisChoice=basisChoice)
    raise AttributeError("wrong version string provided:",version)

#########################################################

# from dune.fem.model import elliptic
from dune.fem.model import integrands
from dune.vem.patch import transform

def vemModel(view, equation, space,
        hessStabilization=None,gradStabilization=None, massStabilization=None,
        *args, **kwargs):
    return integrands(view, equation,
                      modelPatch=transform(space,hessStabilization,gradStabilization,massStabilization),
                      includes=["dune/vem/py/integrands.hh"],
                      *args, **kwargs)

def vemScheme(model, space=None, solver=None, parameters={},
              hessStabilization=None, gradStabilization=None, massStabilization=None,
              boundary="full"):
    """create a scheme for solving second order pdes with the virtual element method

    Args:

    Returns:
        Scheme: the constructed scheme
    """

    from dune.fem.scheme import module
    from dune.fem.scheme import femschemeModule

    assert boundary=="full" or boundary=="value" or boundary=="derivative" or boundary is None

    modelParam = None
    if isinstance(model, (list, tuple)):
        modelParam = model[1:]
        model = model[0]
    if isinstance(model,Equation):
        if space == None:
            try:
                space = model.lhs.arguments()[0].ufl_function_space()
            except AttributeError:
                raise ValueError("no space provided and could not deduce from form provided")
        else:
            try:
                eqSpace = model.lhs.arguments()[0].ufl_function_space()
                if not eqSpace.dimRange == space.dimRange:
                    raise ValueError("""range of space used for arguments in equation
                    and the range of space passed to the scheme have to match -
                    space argument to scheme can be 'None'""")
            except AttributeError:
                pass
        if modelParam:
            model = vemModel(space.grid,model,space,hessStabilization,gradStabilization,massStabilization,*modelParam)
        else:
            model = vemModel(space.grid,model,space,hessStabilization,gradStabilization,massStabilization)

    includes = [ "dune/vem/operator/vemelliptic.hh", "dune/vem/operator/diffusionmodel.hh" ]

    op = lambda linOp,model: "DifferentiableVEMEllipticOperator< " +\
                             ",".join([linOp,model]) + ">"

    if model.hasDirichletBoundary:
        assert boundary is not None
        includes += [ "dune/fem/schemes/dirichletwrapper.hh",
                      "dune/vem/operator/vemdirichletconstraints.hh"]
        if boundary   == "full":       boundary = 3
        elif boundary == "value":      boundary = 1
        elif boundary == "derivative": boundary = 2
        constraints = lambda model: "Dune::VemDirichletConstraints< " +\
                    ",".join([model,space._typeName,str(boundary)]) + " > "
        operator = lambda linOp,model: "DirichletWrapperOperator< " +\
                ",".join([op(linOp,model),constraints(model)]) + " >"
    else:
        assert boundary is None
        operator = op

    spaceType = space._typeName
    # modelType = "VEMDiffusionModel< " +\
    #       "typename " + spaceType + "::GridPartType, " +\
    #       spaceType + "::dimRange, " +\
    #       spaceType + "::dimRange, " +\
    #       "typename " + spaceType + "::RangeFieldType >"

    includes += ["dune/vem/operator/diffusionmodel.hh"]
    valueType = 'std::tuple< typename ' + spaceType + '::RangeType, typename ' + spaceType + '::JacobianRangeType >'
    modelType = 'Dune::Fem::VirtualizedVemIntegrands< typename ' + spaceType + '::GridPartType, ' + model._domainValueType + ", " + model._rangeValueType+ ' >'

    return femschemeModule(space,model,includes,solver,operator,
            parameters=parameters,
            modelType=modelType
            )


def vemOperator(model, domainSpace=None, rangeSpace=None):
    # from dune.fem.model._models import elliptic as makeElliptic
    from dune.fem.operator import load
    if rangeSpace is None:
        rangeSpace = domainSpace

    modelParam = None
    if isinstance(model, (list, tuple)):
        modelParam = model[1:]
        model = model[0]
    if isinstance(model,Form):
        model = model == 0
    if isinstance(model,Equation):
        if rangeSpace == None:
            try:
                rangeSpace = model.lhs.arguments()[0].ufl_function_space()
            except AttributeError:
                raise ValueError("no range space provided and could not deduce from form provided")
        if domainSpace == None:
            try:
                domainSpace = model.lhs.arguments()[1].ufl_function_space()
            except AttributeError:
                raise ValueError("no domain space provided and could not deduce from form provided")
        if modelParam:
            model = vemModel(domainSpace.grid,model,domainSpace,None,None,*modelParam)
        else:
            model = vemModel(domainSpace.grid,model,domainSpace,None,None)

    if not hasattr(rangeSpace,"interpolate"):
        raise ValueError("wrong range space")
    if not hasattr(domainSpace,"interpolate"):
        raise ValueError("wrong domain space")

    domainSpaceType = domainSpace._typeName
    rangeSpaceType = rangeSpace._typeName

    storage,  domainFunctionIncludes, domainFunctionType, _, _, dbackend = domainSpace.storage
    rstorage, rangeFunctionIncludes,  rangeFunctionType,  _, _, rbackend = rangeSpace.storage
    if not rstorage == storage:
        raise ValueError("storage for both spaces must be identical to construct operator")

    includes = ["dune/vem/operator/vemelliptic.hh",
                "dune/fempy/py/grid/gridpart.hh",
                "dune/fem/schemes/dirichletwrapper.hh",
                "dune/vem/operator/vemdirichletconstraints.hh"]
    includes += domainSpace._includes + domainFunctionIncludes
    includes += rangeSpace._includes + rangeFunctionIncludes
    includes += ["dune/vem/operator/diffusionmodel.hh", "dune/fempy/parameter.hh"]

    import dune.create as create
    linearOperator = create.discretefunction(storage)(domainSpace,rangeSpace)[3]

    modelType = "VEMDiffusionModel< " +\
          "typename " + domainSpaceType + "::GridPartType, " +\
          domainSpaceType + "::dimRange, " +\
          rangeSpaceType + "::dimRange, " +\
          "typename " + domainSpaceType + "::RangeFieldType >"
    typeName = "DifferentiableVEMEllipticOperator< " + linearOperator + ", " + modelType + ">"
    if model.hasDirichletBoundary:
        constraints = "Dune::VemDirichletConstraints< " +\
                ",".join([modelType,domainSpace._typeName]) + " > "
        typeName = "DirichletWrapperOperator< " +\
                ",".join([typeName,constraints(model)]) + " >"

    constructor = Constructor(['const '+domainSpaceType+'& dSpace','const '+rangeSpaceType+' &rSpace', modelType + ' &model'],
                              ['return new ' + typeName + '( dSpace, rSpace, model );'],
                              ['pybind11::keep_alive< 1, 2 >()', 'pybind11::keep_alive< 1, 3 >()', 'pybind11::keep_alive< 1, 4 >()'])

    scheme = load(includes, typeName, constructor).Operator(domainSpace,rangeSpace, model)
    scheme.model = model
    return scheme

#################################################################

def aluGrid(constructor, dimgrid=None, dimworld=None, elementType=None, **parameters):
    if not dimgrid:
        dimgrid = getDimgrid(constructor)
    if dimworld is None:
        dimworld = dimgrid
    if elementType is None:
        elementType = parameters.pop("type")
    refinement = parameters["refinement"]
    if refinement == "conforming":
        refinement="Dune::conforming"
    elif refinement == "nonconforming":
        refinement="Dune::nonconforming"
    if not (2 <= dimworld and dimworld <= 3):
        raise KeyError("Parameter error in ALUGrid with dimworld=" + str(dimworld) + ": dimworld has to be either 2 or 3")
    if not (2 <= dimgrid and dimgrid <= dimworld):
        raise KeyError("Parameter error in ALUGrid with dimgrid=" + str(dimgrid) + ": dimgrid has to be either 2 or 3")
    if refinement=="Dune::conforming" and elementType=="Dune::cube":
        raise KeyError("Parameter error in ALUGrid with refinement=" + refinement + " and type=" + elementType + ": conforming refinement is only available with simplex element type")
    typeName = "Dune::ALUGrid< " + str(dimgrid) + ", " + str(dimworld) + ", " + elementType + ", " + refinement + " >"
    includes = ["dune/alugrid/grid.hh", "dune/alugrid/dgf.hh"]
    gridModule = module(includes, typeName, dynamicAttr=True)
    return gridModule.LeafGrid(gridModule.reader(constructor))
def aluSimplexGrid(constructor, dimgrid=2, dimworld=2):
    from dune.grid.grid_generator import module, getDimgrid
    typeName = "Dune::Vem::Grid<"+str(dimgrid)+","+str(dimworld)+">"
    includes = ["dune/vem/misc/grid.hh"]
    gridModule = module(includes, typeName, dynamicAttr=True)
    return gridModule.LeafGrid(gridModule.reader(constructor))
def aluCubeGrid(constructor, dimgrid=2, dimworld=2):
    from dune.grid.grid_generator import module, getDimgrid
    typeName = "Dune::Vem::CubeGrid<"+str(dimgrid)+","+str(dimworld)+">"
    includes = ["dune/vem/misc/grid.hh"]
    gridModule = module(includes, typeName, dynamicAttr=True)
    return gridModule.LeafGrid(gridModule.reader(constructor))
def yaspGrid(constructor, dimgrid=None, coordinates="equidistant", ctype="double"):
    from dune.generator import Constructor
    from dune.grid.grid_generator import module, getDimgrid
    from dune.typeregistry import generateTypeName

    if not dimgrid:
        dimgrid = getDimgrid(constructor)
    if coordinates == "equidistant":
        coordinates = "Dune::EquidistantOffsetCoordinates"
    elif coordinates == "tensorproduct":
        coordinates = "Dune::TensorProductCoordinates"

    if not dimgrid:
        dimgrid = getDimgrid(constructor)
    coordinates, _ = generateTypeName(coordinates, ctype, dimgrid)
    typeName, includes = generateTypeName("Dune::Vem::YGrid", dimgrid, coordinates)
    includes += ["dune/vem/misc/grid.hh", "dune/grid/io/file/dgfparser/dgfyasp.hh"]

    ctor = Constructor([
              'Dune::FieldVector<'+ctype+', '+str(dimgrid)+'> lowerleft',
              'Dune::FieldVector<'+ctype+', '+str(dimgrid)+'> upperright',
              'std::array<int, '+str(dimgrid)+'> elements',
              'std::array<bool, '+str(dimgrid)+'> periodic',
              'int overlap'],
              ['std::bitset<'+str(dimgrid)+'> periodic_;',
               'for (int i=0;i<'+str(dimgrid)+';++i) periodic_.set(i,periodic[i]);',
               'return new DuneType(lowerleft,upperright,elements,periodic_,overlap);'],
              ['"lowerleft"_a', '"upperright"_a', '"elements"_a', '"periodic"_a', '"overlap"_a'])

    gridModule = module(includes, typeName, ctor, dynamicAttr=True)

    try:
        lowerleft  = constructor.lower
        upperright = constructor.upper
        elements   = constructor.division
        periodic   = constructor.param.get("periodic", [False,]*dimgrid)
        overlap    = constructor.param.get("overlap", 0)
        return gridModule.HierarchicalGrid(lowerleft,upperright,elements,periodic,overlap).leafView
    except AttributeError:
        return gridModule.reader(constructor).leafView

#################################################################
class TrivialAgglomerate:
    def __init__(self,constructor, cubes=False, **kwargs):
        if cubes:
            self.grid = aluCubeGrid(constructor, **kwargs)
        else:
            self.grid = aluSimplexGrid(constructor, **kwargs)
        self.suffix = "simple"+str(self.grid.size(0))
        self.size = self.grid.size(0)
    def __call__(self,en):
        return self.grid.indexSet.index(en)
    def check(self):
        return True
# http://zderadicka.eu/voronoi-diagrams/
from dune.vem.voronoi import triangulated_voronoi
from scipy.spatial import Voronoi, voronoi_plot_2d, cKDTree, Delaunay
import numpy

from sortedcontainers import SortedDict
import triangle
import matplotlib.pyplot as plt
class PolyAgglomerate:
    def __init__(self,constructor,cubes=False,convex=False):
        self.convex = convex
        self.cubes = cubes
        self.suffix = "poly"+str(len(constructor["polygons"]))
        self.domain, self.index = self.construct(constructor["vertices"],
                                                 constructor["polygons"])
        if cubes:
            self.grid = aluCubeGrid(self.domain)
        else:
            self.grid = aluSimplexGrid(self.domain)
        self.ind = set()
        self.size = len( constructor["polygons"] )
    def __call__(self,en):
        bary = en.geometry.center
        return self.index[self.roundBary(bary)]
    def check(self):
        return len(self.ind)==self.N
    def roundBary(self,a):
        return tuple(round(aa,8) for aa in a)
    def construct(self,vertices,polygons):
        index = SortedDict()
        if self.cubes:
            for i,poly in enumerate(polygons):
                assert len(poly)==4
                bary = [sum([vertices[p][i] for p in poly]) / 4 for i in range(2)]
                index[self.roundBary(bary)] = i
                polygons[i][2], polygons[i][3] = poly[3], poly[2]
            domain = {"vertices": vertices, "cubes": polygons}
        else:
            vertices = numpy.array(vertices)
            tr = []
            for nr, p in enumerate(polygons):
                p = numpy.array(p)
                N = len(p)
                if not self.convex: # use triangle
                    e = [ [p[i],p[(i+1)%N]] for i in range(N) ]
                    domain = { "vertices":vertices,
                               "segments":numpy.array(e) }
                    tr += [triangle.triangulate(domain,opts="p")]
                else: # use scipy
                    poly = numpy.append(p,[p[0]])
                    vert = vertices[p, :]
                    tri = Delaunay(vert).simplices
                    tr += [{"triangles":p[tri]}]
                bary = [ (vertices[p0]+vertices[p1]+vertices[p2])/3.
                         for p0,p1,p2 in tr[-1]["triangles"] ]
                for b in bary:
                    index[self.roundBary(b)] = nr
            domain = {"vertices":  numpy.array(vertices),
                      "simplices": numpy.vstack([ t["triangles"] for t in tr ])}
        return domain, index

def polyGrid(constructor,cubes=False, convex=False,**kwargs):
    if isinstance(constructor,dict) and \
        constructor.get("polygons",None) is not None:
        agglomerate = PolyAgglomerate(constructor,cubes,convex)
        agglomerate.minEdgeNumber = min([len(p) for p in constructor["polygons"]])
    else:
        agglomerate = TrivialAgglomerate(constructor, cubes, **kwargs)
        if cubes:
            agglomerate.minEdgeNumber = 4
        else:
            agglomerate.minEdgeNumber = 3
    grid = agglomerate.grid
    grid.hierarchicalGrid.agglomerate = agglomerate
    return grid



import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import math
from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate, discreteFunction
from dune.fem import parameter
from dune.vem import voronoiCells
from dune.fem.operator import linear as linearOperator
from scipy.sparse.linalg import spsolve

from ufl import *
import dune.ufl

maxLevel     = 5
order        = 5
epsilon      = 1
laplaceCoeff = 1
mu           = 0

dune.fem.parameter.append({"fem.verboserank": 0})

# <markdowncell>
# we can compare different method, e.g., a lagrange space (on the # subtriangulation),
# a bounding box dg space and a conforming/non conforming VEM space

# <codecell>
# Note: suboptimal laplace error for bubble (space is reduced to polorder=3 but could be 4 = ts+2
methods = [ ### "[space,scheme,spaceKwrags]"
        # ["vem","vem",{"order":order, "testSpaces":[ [0],  [order-3,order-2], [order-4] ] }, "C1-non-conforming"],
        # ["vem","vem",{"order":order, "testSpaces":[ [0],  [order-2,order-2], [order-2] ] }, "C1C0-conforming"],
        ["vem","vem",{"order":order, "testSpaces":[ [0],  [order-3,order-2], [order-3] ] }, "C1mod-conforming"],
          ]
if epsilon == 0:
    methods += [
            ["vem","vem",{"order":order, "testSpaces":[ [0],  [order-2,-1], [order-2] ] },      "C0-conforming"],
            ["vem","vem",{"order":order-1, "testSpaces":[ [0],  [order-3,-1], [order-3] ] },      "C0pm1-conforming"],
               ]
parameters = {"newton.linear.tolerance": 1e-12,
              "newton.linear.preconditioning.method": "jacobi",
              "penalty": 40,  # for the bbdg scheme
              "newton.linear.verbose": False,
              "newton.verbose": True
              }

# <markdowncell>
# Now we define the model starting with the exact solution:

# <codecell>
uflSpace = dune.ufl.Space(2, dimRange=1)
x = SpatialCoordinate(uflSpace)
exact = as_vector( [sin(2*pi*x[0])**2*sin(2*pi*x[1])**2] )

# next the bilinear form
# Note: for function which continuous derivatives we have
#       laplace(u)*laplace(v)*dx = inner(u,v)*dx
#       as can be seen by using integration by parts on the mixed terms on the right
#       and using continuety of u,v
#       For the non conforming spaces we don't have continuety of the
#       derivatives so the equivalence does not hold and one should use the
#       right hand side directly to obtain a coercive bilinear form w.r.t.
#       the norm on H^2 (the left is not a norm in this case).
#       For computing the forcing term 'b' both formula are fine since
#       'exact' is smooth enough.
laplace = lambda w: div(grad(w))
H = lambda w: grad(grad(w))
u = TrialFunction(uflSpace)
v = TestFunction(uflSpace)
a = ( epsilon*inner(H(u[0]),H(v[0])) +\
      laplaceCoeff*inner(grad(u),grad(v)) +\
      mu*inner(u,v)
    ) * dx

# finally the right hand side and the boundary conditions
b = ( epsilon*laplace(laplace(exact[0])) -\
      laplaceCoeff*laplace(exact[0]) +\
      mu*exact[0] ) * v[0] * dx
dbc = [dune.ufl.DirichletBC(uflSpace, [0], i+1) for i in range(4)]
biLaplaceCoeff = epsilon
diffCoeff      = laplaceCoeff
massCoeff      = mu

# <markdowncell>
# Now we define a grid build up of voronoi cells around $50$ random points

# We now define a function to compute the solution and the $L^2,H^1$ error
# given a grid and a space

# <codecell>
def compute(grid, space, schemeName):
    # solve the pde
    df = discreteFunction(space, name="solution") # space.interpolate([0],name="solution")
    # df.plot(level=3)
    info = {"linear_iterations":1}
    scheme = create.scheme(schemeName, [a==b, *dbc], space,
                        # solver="cg",
                        # solver=("suitesparse","umfpack"),
                        hessStabilization=biLaplaceCoeff,
                        gradStabilization=diffCoeff,
                        massStabilization=massCoeff,
                        parameters=parameters)
    # df.interpolate(exact)
    # info = scheme.solve(target=df)
    jacobian = linearOperator(scheme)
    rhs = discreteFunction(space,name="rhs")
    scheme(df,rhs)
    rhs.as_numpy[:] *= -1
    df.as_numpy[:] = spsolve(jacobian.as_numpy, rhs.as_numpy[:])
    edf = exact-df
    err = [inner(edf,edf),
           inner(grad(edf),grad(edf)),
           inner(grad(grad(edf)),grad(grad(edf)))]
    errors = [ math.sqrt(e) for e in integrate(grid, err, order=8) ]
    return df, errors, info

# <markdowncell>
# Finally we iterate over the requested methods and solve the problems

# <codecell>
# fig = pyplot.figure(figsize=(10*maxLevel,10*len(methods)))
# figPos = 100*len(methods)+10*maxLevel+1
results = []
for level in range(maxLevel):
    constructor = cartesianDomain([0,0],[1,1],[4*2**level,4*2**level])
    polyGrid = create.grid("agglomerate", constructor, cubes=False )
    # constructor = cartesianDomain([0,0],[1,1],[4,4])
    # polyGrid = create.grid("agglomerate", voronoiCells(constructor,16*2**level*2**level,"voronoiseeds",load=True,show=False,lloyd=5) )
    res = []
    for i,m in enumerate(methods):
        space = create.space(m[0], polyGrid, dimRange=1, storage="fem", **m[2])
        dfs,errors,info = compute(polyGrid, space, m[1])
        print("[",level,"]","method:(",m[0],m[2],")",
              "Size: ",space.size, polyGrid.size(0), "L^2: ", errors[0], "H^1: ", errors[1],
              "H^2: ", errors[2],
              info["linear_iterations"], flush=True)
        res += [ [polyGrid.hierarchicalGrid.agglomerate.size,space.size,errors[0],errors[1],errors[2]] ]
        # dfs.plot(level=4)
        # plot(grad(grad(dfs))[0,0,0],grid=polyGrid,level=3,
        # plot(dfs,grid=polyGrid,level=3,
        #      figure=(fig,figPos+i*maxLevel+level),gridLines=None, colorbar="horizontal")
        # interpol = space.interpolate(exact,name="interpolation")
        # plot(grad(grad(interpol))[0,0,0],grid=polyGrid,level=3,
        #      figure=(fig,figPos+level*len(methods)+i+1),gridLines=None, colorbar="horizontal")
    results += [res]
h = lambda size: 1/sqrt(size)
print("# p="+str(order)+", eps="+str(epsilon)+", laplace="+str(laplaceCoeff))
for i,m in enumerate(methods):
    eoc = [-1,-1,-1]
    for level in range(0,maxLevel):
        print(h(results[level][i][0]),*results[level][i],*eoc)
        if level<maxLevel-1:
            eoc = [math.log(results[level+1][i][2+j]/results[level][i][2+j])/
                   math.log(h(results[level+1][i][0])/h(results[level][i][0]))
                   for j in range(3)]
    print("\n")
# pyplot.show()



import dune.vem
import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import math, pickle
from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate
from dune.fem import parameter
from dune.vem import voronoiCells

from ufl import *
import dune.ufl
dune.fem.parameter.append({"fem.verboserank": 0})

# use some run time parameters here to enable batch processesing
# the dump file should include the parameter values and the plot script
# should also use the parameters to determin input and output file names
order = 2
gridTypes = ["triangles","cartesian","quadrilaterals","voronoi","referenceTriangle"]
useGrid = 3 # 0..4
# note that the hessian of the reference element based basis
# function is not implemented for the 'quadrilaterals' grid

# Note: suboptimal laplace error for bubble (space is reduced to polorder=3 but could be 4 = ts+2
methods = [ ### "[space,scheme,spaceKwrags]"
            # ["lagrange","galerkin",{}, "Lagrange"],
            # ["vem","vem",{"testSpaces":[ [0],  [order-2], [order-1] ] }, "Bubble"],
            # ["vem","vem",{"testSpaces":[ [0],  [order-2], [order-3] ] }, "Serendipity"],
            # ["vem","vem",{"testSpaces":[ [-1], [order-1], [order-3] ] }, "Nc-Serendipity"],
            # ["vem","vem",{"conforming":True}, "conforming"],
            # ["vem","vem",{"conforming":False}, "non-conforming"],
            ["vem","vem",{"testSpaces":[ [0],  [order-3,order-2], [order-4] ] }, "C1-non-conforming"],
            ["vem","vem",{"testSpaces":[ [0],  [order-2,order-2], [order-2] ] }, "C1C0-conforming"],
            # ["bbdg","bbdg",{},"bbdg"],
   ]

uflSpace = dune.ufl.Space(2, dimRange=1)
x = SpatialCoordinate(uflSpace)
# exact = as_vector( [ 10 + x[0]*x[1] + x[0]**2*x[1] + x[1]**3 ] )
exact = as_vector( [x[0]*x[1] * cos(pi*x[0]*x[1])] )
# exact = as_vector( [x[0]*x[0]] )
def compute(grid, space, schemeName):
    # do the interpolation
    df = space.interpolate(exact,name="solution")
    info = {"linear_iterations":-1}
    # compute the error
    edf = exact-df
    err = [inner(edf,edf),
           inner(grad(edf),grad(edf)),
           inner(div(grad(edf)),div(grad(edf))),
           inner(grad(grad(edf)),grad(grad(edf)))
           ]
    errors = [ math.sqrt(e) for e in integrate(grid, err, order=8) ]
    if 0:
        plot(grad(df[0])[0],grid=grid,level=3)
        from dune.ufl import CoordWrapper
        import numpy
        df.interpolate([x[0]])
        for d in df.dofVector:
            print(d,end=", ")
        print()
        df.interpolate([x[1]])
        for d in df.dofVector:
            print(d,end=", ")
        print()
        for d in range(len(df.dofVector)):
            print(":::::::::::::::::::::::::::::")
            df.clear()
            df.dofVector[d] = 1
            for el in grid.elements:
                e = CoordWrapper(el,[0,0])
                print(numpy.round(df.ufl_evaluate(e,[0],[]),3),end="  ")
                e = CoordWrapper(el,[1,0])
                print(numpy.round(df.ufl_evaluate(e,[0],[]),3),end="  ")
                e = CoordWrapper(el,[0,1])
                print(numpy.round(df.ufl_evaluate(e,[0],[]),3),end="  |  ")

                e = CoordWrapper(el,[0.5,0])
                g = [ df.ufl_evaluate(e,[0],[0]), df.ufl_evaluate(e,[0],[1]) ]
                print(numpy.round(g[1],3),",",numpy.round(g[0],3),end="  ")
                e = CoordWrapper(el,[0,0.5])
                g = [ df.ufl_evaluate(e,[0],[0]), df.ufl_evaluate(e,[0],[1]) ]
                print(numpy.round(g[1],3),",",numpy.round(g[0],3),end="  ")
                e = CoordWrapper(el,[0.5,0.5])
                g = [ df.ufl_evaluate(e,[0],[0]), df.ufl_evaluate(e,[0],[1]) ]
                print(numpy.round(g[1],3),",",numpy.round(g[0],3),end="  ")
                print(";;;")
            # print(d)
            # print(":::::::::::::::::::::::::::::")
            # print("value")
            # df.plot(level=3)
            # print("gradient")
            # plot(grad(df[0])[0],grid=grid,level=3)
    return df, [ ["L^2",errors[0]],
                 ["H^1",errors[1]],
                 ["laplace",errors[2]],
                 ["H^2",errors[3]],
           ], info

results = []
for level in range(0,4):
    N = 2**(level+2)
    if useGrid == 0:
        constructor = cartesianDomain([-0.5,-0.5],[1,1],[N,N])
        polyGrid = create.grid("agglomerate", constructor, cubes=False )
    elif useGrid == 1:
        constructor = cartesianDomain([-0.5,-0.5],[1,1],[N,N])
        polyGrid = create.grid("agglomerate", constructor, cubes=True )
    elif useGrid == 2:
        # the following is a non affine cube grid, i.e., not parallelograms,
        cubeGrid = {"vertices": [ [-0.5,-0.5],[0   ,-0.5],[1,-0.5],
                                  [ 1,   0],  [1   , 1],
                                  [ 0,   1],  [-0.5, 1],
                                  [-0.5, 0],
                                  [0.25,0.25] ],
                    "cubes":  [ [0,1,7,8],[1,2,8,3],[8,3,5,4],[7,8,6,5] ]}
        polyGrid = create.grid("agglomerate", cubeGrid, cubes=True )
        polyGrid.hierarchicalGrid.globalRefine(level)
    elif useGrid == 3:
        constructor = cartesianDomain([0,0],[1,1],[N,N])
        polyGrid = create.grid("agglomerate",
                  voronoiCells(constructor,N*N,"voronoiseeds",load=True,show=False,lloyd=5) )
    elif useGrid == 4:
        if 1:
            # reference triangle
            refGrid = {"vertices": [ [0,0],[1,0],[0,1] ], "simplices":  [ [0,1,2] ]}
            # refGrid = {"vertices": [ [0,0],[1,0],[1,1],[0,1] ], "polygons": [ [0,1,2,3] ]}
            polyGrid = create.grid("agglomerate", refGrid, cubes=False )
        else:
            refGrid = {"vertices": [ [0,0],[1,0],[0,1],[1,1] ], "cubes": [ [0,1,2,3] ]}
            polyGrid = create.grid("agglomerate", refGrid, cubes=True )
        polyGrid.hierarchicalGrid.globalRefine(level)
    gridVidth = -1 # need something here
    @gridFunction(polyGrid, name="cells")
    def polygons(en,x):
        return polyGrid.hierarchicalGrid.agglomerate(en)
    # polygons.plot(colorbar="horizontal")

    # figCols = 4
    # figRows = len(methods)//figCols+1
    # fig = pyplot.figure(figsize=(10*figCols,10*figRows))
    # figPos = 100*figRows+10*figCols+1
    res = []
    for i,m in enumerate(methods):
        # dune.generator.setFlags(flags="-g",noChecks=True)
        space = create.space(m[0], polyGrid, order=order, dimRange=1, storage="istl", **m[2])
        # dune.generator.unsetFlags()
        dfs,errors,info = compute(polyGrid, space, m[1])
        print("method:(",m[0],m[2],")",
              "Size: ",space.size,
              *[e for e in errors],
              flush=True)
        # dfs.plot(figure=(fig,figPos+i),gridLines=None, colorbar="horizontal",level=3)
        # plot(grad(dfs)[0,0],grid=polyGrid,level=3,
        #      figure=(fig,figPos+i),gridLines=None, colorbar="horizontal")
        # plot(grad(grad(dfs))[0,0,0],grid=polyGrid,level=3,
        #      figure=(fig,figPos+i),gridLines=None, colorbar="horizontal")
        res += [ [m, gridVidth, space.size, errors] ]

    results += [res]
    pickle.dump(results,open('interpolation.dump','wb'))
print(results)
# pyplot.show()



# http://stackoverflow.com/questions/28665491/getting-a-bounded-polygon-coordinates-from-voronoi-cells

import matplotlib.pyplot as pl
import numpy as np
import scipy as sp
import scipy.spatial
import sys, os, pickle
from scipy.spatial import Voronoi, voronoi_plot_2d, cKDTree
import numpy

eps = sys.float_info.epsilon

def in_box(towers, bounding_box):
    return np.logical_and(np.logical_and(bounding_box[0] <= towers[:, 0],
                                         towers[:, 0] <= bounding_box[1]),
                          np.logical_and(bounding_box[2] <= towers[:, 1],
                                         towers[:, 1] <= bounding_box[3]))

def centroid(vertices):
    signed_area = 0
    C_x = 0
    C_y = 0
    for i in range(len(vertices)):
        ip = (i+1) % len(vertices)
        step = (vertices[i, 0]*vertices[ip, 1])-(vertices[ip, 0]*vertices[i, 1])
        signed_area += step
        C_x += (vertices[i, 0] + vertices[ip, 0])*step
        C_y += (vertices[i, 1] + vertices[ip, 1])*step
    signed_area = 1/2*signed_area
    C_x = (1.0/(6.0*signed_area))*C_x
    C_y = (1.0/(6.0*signed_area))*C_y
    return np.array([C_x, C_y])

def voronoiCells(constructor, towers, fileName=None, load=False, lloyd=False, show=False):
    try:
        constructor = [constructor.lower,constructor.upper]
    except AttributeError:
        pass
    lowerleft  = numpy.array(constructor[0])
    upperright = numpy.array(constructor[1])
    bounding_box = numpy.array(
            [lowerleft[0],upperright[0],lowerleft[1],upperright[1]] )

    if isinstance(towers,int):
        if fileName:
            fileName = fileName + str(towers) + '.pickle'
            if not load or not os.path.exists(fileName):
                # print("generating new seeds for voronoi grid")
                numpy.random.seed(1234)
                towers = numpy.array(
                        [ p*(upperright-lowerleft) + lowerleft
                            for p in numpy.random.rand(towers, 2) ])
                with open(fileName, 'wb') as f:
                    pickle.dump(towers, f)
            else:
                # print("loading seeds for voronoi grid")
                with open(fileName, 'rb') as f:
                    towers = pickle.load(f)
        else:
            towers = numpy.array(
                    [ p*(upperright-lowerleft) + lowerleft
                        for p in numpy.random.rand(towers, 2) ])
    towers = numpy.array(towers)

    # Select towers inside the bounding box
    i = in_box(towers, bounding_box)

    vor = sp.spatial.Voronoi(towers[i,:],incremental=True)
    if show:
        if lloyd is not False:
            voronoi_plot_2d(vor,show_points=False,show_vertices=False).\
                savefig("lloyd"+str(lloyd)+"inc"+str(len(towers))+".pdf", bbox_inches='tight')
        else:
            voronoi_plot_2d(vor,show_points=False,show_vertices=False).\
                savefig(fileName+"inc"+str(len(towers))+".pdf", bbox_inches='tight')

    # Mirror points
    points_center = towers[i, :]
    points_left = np.copy(points_center)
    points_left[:, 0] = bounding_box[0] - (points_left[:, 0] - bounding_box[0])
    points_right = np.copy(points_center)
    points_right[:, 0] = bounding_box[1] + (bounding_box[1] - points_right[:, 0])
    points_down = np.copy(points_center)
    points_down[:, 1] = bounding_box[2] - (points_down[:, 1] - bounding_box[2])
    points_up = np.copy(points_center)
    points_up[:, 1] = bounding_box[3] + (bounding_box[3] - points_up[:, 1])
    points = np.append(points_center,
                       np.append(np.append(points_left,
                                           points_right,
                                           axis=0),
                                 np.append(points_down,
                                           points_up,
                                           axis=0),
                                 axis=0),
                       axis=0)
    # Compute Voronoi
    vor = sp.spatial.Voronoi(points)

    # Filter regions
    regions = []
    for i,region in enumerate(vor.regions):
        flag = True
        for index in region:
            if index == -1:
                flag = False
                break
            else:
                x = vor.vertices[index, 0]
                y = vor.vertices[index, 1]
                if not(bounding_box[0] - eps <= x and x <= bounding_box[1] + eps and
                       bounding_box[2] - eps <= y and y <= bounding_box[3] + eps):
                    flag = False
                    break
        if region != [] and flag:
            regions.append(region)

    if lloyd > 0:
        dist = 0
        seeds = []
        for i,r in enumerate(regions):
            # old = seeds[i].copy()
            # new = centroid(vor.vertices[r])
            # dist = max(dist,numpy.linalg.norm(new-old))
            seeds.append(centroid(vor.vertices[r]))
        if type(lloyd) == int:
            return voronoiCells(constructor, seeds, fileName=None, load=False, lloyd=lloyd-1, show=show)
        elif dist > lloyd:
            return voronoiCells(constructor, towers, fileName=None, load=False, lloyd=lloyd, show=show)

    lowerleft  = numpy.array(constructor[0])
    upperright = numpy.array(constructor[1])
    bounding_box = numpy.array(
            [lowerleft[0],upperright[0],lowerleft[1],upperright[1]] )

    indices = set()
    for poly in regions:
        for r in poly:
            indices.add( r )
    indices = np.array(list(indices))
    vorVertices = vor.vertices[indices,:]
    newind = np.zeros(len(vor.vertices),int)
    for i in range(len(indices)):
        newind[indices[i]] = i

    return {"vertices":vorVertices, "polygons":[newind[r] for r in regions]}

def triangulated_voronoi(constructor, towers):
    voronoi = voronoiCells(constructor, towers)
    vorVertices, polys = voronoi["vertices"], voronoi["polygons"]
    indices = set()
    triangles = np.zeros([0,3],int)
    minEdgeNumber = 1000
    for poly in polys:
        p = numpy.append(poly,[poly[0]])
        vert = vorVertices[p, :]
        for r in poly:
            indices.add( r )
        tri = sp.spatial.Delaunay(vert).simplices
        triangles = np.concatenate(
                (triangles, p[tri] ),
                axis=0 )
        minEdgeNumber = min(minEdgeNumber,len(p))

    indices = np.array(list(indices))
    vertices = vorVertices[indices,:]
    newind = np.zeros(len(vorVertices),int)
    for i in range(len(indices)):
        newind[indices[i]] = i
    return vertices, newind[triangles],minEdgeNumber



# n_towers = 100
# towers = np.random.rand(n_towers, 2)
# bounding_box = np.array([0., 1., 0., 1.]) # [x_min, x_max, y_min, y_max]

# points, triangles = triangulated_voronoi(towers, bounding_box)
# pl.triplot(points[:,0], points[:,1], triangles.copy())
# pl.plot(points[:,0], points[:,1], 'o')
# pl.show()



import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot

from ufl import SpatialCoordinate, CellVolume, TrialFunction, TestFunction,\
                inner, dot, div, grad, dx, as_vector, transpose, Identity
from dune.ufl import Constant, DirichletBC

from dune.fem.operator import linear as linearOperator
from dune.fem import parameter
parameter.append({"fem.verboserank": 0})

useCartesian  = False
uzawaPreconditioner = True
storage="petsc"

mu_   = 0.1
nu_   = 10.0

def compute(useVem,order,version,plot=False):
    print("########################################################")
    print(useVem,order,version)
    print("########################################################")
    # grid construction
    from dune.grid import structuredGrid, cartesianDomain
    from dune.vem import polyGrid        as leafGridView
    domain = cartesianDomain([0,0],[3,1],[30,10])
    if useCartesian:
        constructor = domain
    else:
        from dune.vem import voronoiCells
        constructor = voronoiCells(domain,300,"uzawaseeds",True)
    grid = leafGridView( constructor )
    if plot: grid.plot()

    # spaces and schemes
    from dune.fem.operator import galerkin   as galerkinOperator
    if not useVem:
        from dune.fem.space    import lagrange   as velocitySpace
        from dune.fem.scheme   import galerkin   as velocityScheme
        if version=="TH":
            from dune.fem.space  import lagrange as pressureSpace
            from dune.fem.scheme import galerkin as pressureScheme
            useDGPressure = False
        else:
            from dune.fem.space  import dgonb    as pressureSpace
            from dune.fem.scheme import dg       as pressureScheme
            useDGPressure = True
        spcU = velocitySpace(grid, dimRange=grid.dimension, order=order, storage=storage)
        spcP = pressureSpace(grid, dimRange=1, order=order-1, storage=storage),
    else:
        if version=="TH":
            veloTestSpaces = [order,   [0,order-2,order-2]]
            presTestSpaces = [order-1, [0,order-3,order-3]]
            useDGPressure = False
        if version=="mini":
            veloTestSpaces = [order,   [0,order-2,order-1]]
            presTestSpaces = [order,   [0,order-2,order-2]]
            useDGPressure = False
        if version=="non-conf":
            veloTestSpaces = [order,   [-1,order-1,order-2]]
            presTestSpaces = order-1
            useDGPressure = True
        if version=="SV":
            veloTestSpaces = [order,   [0,order-2,order-2]]
            presTestSpaces = order-1
            useDGPressure = True
        from dune.vem import vemSpace   as velocitySpace
        from dune.vem import vemScheme  as velocityScheme
        if not useDGPressure:
            from dune.vem import vemSpace   as pressureSpace
            from dune.vem import vemScheme  as pressureScheme
            spcU = velocitySpace(grid, dimRange=grid.dimension, storage=storage,
                      order=veloTestSpaces[0], testSpaces=veloTestSpaces[1])
            spcP = pressureSpace(grid, dimRange=1, storage=storage,
                      order=presTestSpaces[0], testSpaces=presTestSpaces[1])
        else:
            from dune.vem import bbdgSpace  as pressureSpace
            from dune.vem import bbdgScheme as pressureScheme
            spcU = velocitySpace(grid, dimRange=grid.dimension, storage=storage,
                      order=veloTestSpaces[0], testSpaces=veloTestSpaces[1])
            spcP = pressureSpace(grid, dimRange=1, storage=storage,
                      order=presTestSpaces)

    velocity = spcU.interpolate([0,]*spcU.dimRange, name="velocity")
    pressure = spcP.interpolate([0], name="pressure")
    print("size of pressure space:",spcP.size)
    print("size of velocity space:",spcU.size)

    # setting up the problem
    cell  = spcU.cell()
    x     = SpatialCoordinate(cell)
    mu    = Constant(mu_,  "mu")
    nu    = Constant(nu_,  "nu")
    u     = TrialFunction(spcU)
    v     = TestFunction(spcU)
    p     = TrialFunction(spcP)
    q     = TestFunction(spcP)

    # exact solution and resulting forcing
    exact_u     = as_vector( [x[1] * (1.-x[1]), 0] )
    exact_p     = as_vector( [ (-2*x[0] + 2)*mu ] )
    f           = as_vector( [0,]*grid.dimension )
    f          += nu*exact_u

    # ufl forms
    mainModel   = (nu*dot(u,v) + mu*inner(grad(u)+grad(u).T, grad(v)) - dot(f,v)) * dx
    gradModel   = -inner( p[0]*Identity(grid.dimension), grad(v) ) * dx
    divModel    = -div(u)*q[0] * dx
    massModel   = inner(p,q) * dx
    preconModel = inner(grad(p),grad(q)) * dx

    # operators and schemes
    if not useVem:
        mainOp       = velocityScheme([mainModel==0,DirichletBC(spcU,exact_u,1)])
        if useDGPressure:
            preconOp = pressureScheme(preconModel==0, parameters={"penalty":20})
        else:
            preconOp = pressureScheme(preconModel==0)
    else:
        mainOp       = velocityScheme([mainModel==0,DirichletBC(spcU,exact_u,1)],
                             gradStabilization=as_vector([2*mu_,2*mu_]),
                             massStabilization=as_vector([nu_,nu_]) )
        if useDGPressure:
            preconOp = pressureScheme(preconModel==0, parameters={"penalty":20})
        else:
            preconOp = pressureScheme(preconModel==0, gradStabilization=1)
    mainOp.model.mu = mu_
    mainOp.model.nu = nu_

    gradOp      = galerkinOperator(gradModel)
    divOp       = galerkinOperator(divModel)
    massOp      = pressureScheme(massModel==0)

    A = linearOperator(mainOp)
    G = linearOperator(gradOp)
    D = linearOperator(divOp)
    M = linearOperator(massOp)
    P = linearOperator(preconOp)

    solver = {"method":"cg","tolerance":1e-10, "verbose":False}
    Ainv   = mainOp.inverseLinearOperator(A,parameters=solver)
    Minv   = massOp.inverseLinearOperator(M,parameters=solver)
    Pinv   = preconOp.inverseLinearOperator(P,solver)

    # auxiliary variables for uzawa algorithm
    rhsVelo  = velocity.copy()
    rhsPress = pressure.copy()
    r      = rhsPress.copy()
    d      = rhsPress.copy()
    precon = rhsPress.copy()
    xi     = rhsVelo.copy()

    # compute initial residual
    mainOp(velocity,rhsVelo)
    rhsVelo *= -1
    G(pressure,xi)
    rhsVelo -= xi
    mainOp.setConstraints(rhsVelo)
    mainOp.setConstraints(velocity)

    Ainv(rhsVelo, velocity)
    D(velocity,rhsPress)
    Minv(rhsPress, r)

    if uzawaPreconditioner and mainOp.model.nu > 0:
        precon.clear()
        Pinv(rhsPress, precon)
        r *= mainOp.model.mu
        r.axpy(mainOp.model.nu,precon)

    d.assign(r)
    delta = r.scalarProductDofs(rhsPress)
    print("delta:",delta,flush=True)
    assert delta >= 0

    # start iteration
    for m in range(100):
        xi.clear()
        G(d,rhsVelo)
        mainOp.setConstraints(\
            [0,]*grid.dimension, rhsVelo)
        Ainv(rhsVelo, xi)
        D(xi,rhsPress)
        rho = delta /\
           d.scalarProductDofs(rhsPress)
        pressure.axpy(rho,d)
        velocity.axpy(-rho,xi)
        D(velocity, rhsPress)
        Minv(rhsPress,r)
        if uzawaPreconditioner and mainOp.model.nu > 0:
            precon.clear()
            Pinv(rhsPress,precon)
            r *= mainOp.model.mu
            r.axpy(mainOp.model.nu,precon)
        oldDelta = delta
        delta = r.scalarProductDofs(rhsPress)
        print("delta:",delta,flush=True)
        if delta < 1e-9: break
        gamma = delta/oldDelta
        d *= gamma
        d += r
    if plot:
        velocity.plot(colorbar="horizontal")
        pressure.plot(colorbar="horizontal")
    return velocity,pressure

fig = pyplot.figure(figsize=(40,20))
order = 2
solutions = []
solutions += [compute(True, order, "TH")]
solutions[0][0].plot(figure=(fig,421),gridLines=None, colorbar=None)
solutions[0][1].plot(figure=(fig,422),gridLines=None, colorbar=None)
solutions += [compute(True, order-1, "mini")]
solutions[1][0].plot(figure=(fig,423),gridLines=None, colorbar=None)
solutions[1][1].plot(figure=(fig,424),gridLines=None, colorbar=None)
solutions += [compute(True, order, "non-conf")]
solutions[2][0].plot(figure=(fig,425),gridLines=None, colorbar=None)
solutions[2][1].plot(figure=(fig,426),gridLines=None, colorbar=None)
solutions += [compute(True, order, "SV")]
solutions[3][0].plot(figure=(fig,427),gridLines=None, colorbar=None)
solutions[3][1].plot(figure=(fig,428),gridLines=None, colorbar=None)
pyplot.tight_layout()
pyplot.show()



import pickle
from matplotlib import pyplot
import numpy
from math import log

results = pickle.load(open('interpolation.dump','rb'))
# [ [ m, gridVidth, space.size, errors ], [...],... ]
# m = [*,*,testSpaces]
# errors = [[name,err],...]

# get different gridWidth and space size values, method and error names
gw = []
for r in results:
    gw.append(r[0][1])
    numberOfErrors  = len(r[0][3])
    methods = []
    for m in r:
        methods.append(m[0][3])
    errorNames  = []
    for e in r[0][3]:
        errorNames.append(e[0])
dofs   = numpy.zeros((len(methods),len(gw)))
errors = numpy.zeros((len(methods),len(gw),numberOfErrors))
eocs   = numpy.zeros((len(methods),len(gw)-1,numberOfErrors))
print(errors.shape)

for i,r in enumerate(results):
    for j,m in enumerate(r):
        dofs[j,i]     = m[2]
        errors[j,i,:] = [i[1] for i in m[3]]
for j in range( errors.shape[0] ):
    for i in range( errors.shape[2] ):
        eocs[j,:,i] = numpy.log( errors[j,1:,i] / errors[j,:-1,i] ) / numpy.log(0.5)

# the following should be improved on...
markers = ['x-', 's-', '+-', 'o-', '<-', '>-', '*-']
colors  = ['coral', 'skyblue', 'indianred', 'blueviolet', 'green']

for e, errName in enumerate(errorNames):
    pyplot.figure()
    pyplot.ylabel(errName+' error',fontsize=20)
    for i in range(dofs.shape[0]):
        pyplot.loglog(dofs[i],errors[i,:,e],
                 markers[i%7], color=colors[i%5], markersize=15, label=methods[i])
    pyplot.grid(True)
    pyplot.tick_params(axis='both', which='minor', labelsize=32)
    pyplot.xlabel('number of dofs',fontsize=20)
    pyplot.legend(loc="lower left")
    pyplot.savefig('interpolationError'+errName+'.eps')
    pyplot.show()

    pyplot.figure()
    pyplot.ylabel(errName+' eoc',fontsize=20)
    for i in range(dofs.shape[0]):
        pyplot.semilogx(dofs[i,1:],eocs[i,:,e],
                 markers[i%7], color=colors[i%5], markersize=15, label=methods[i])
    pyplot.grid(True)
    pyplot.tick_params(axis='both', which='minor', labelsize=32)
    pyplot.xlabel('number of dofs',fontsize=20)
    pyplot.legend(loc="lower left")
    pyplot.savefig('interpolationEoc'+errName+'.eps')
    pyplot.show()



# <markdowncell>
# # Demonstration notebook for the DUNE-VEM module
# This module is based on DUNE-FEM # (https://www.dune-project.org/modules/dune-fem)
# and provides a Python frontend based on the new Python extension to DUNE
# (https://dune-project.org/modules/dune-python)
#
# ## Laplace problem
#
# We first consider a simple Laplace problem with Dirichlet boundary conditions
# \begin{align*}
#   -\Delta u &= f, && \text{in } \Omega, \\
#           u &= g, && \text{on } \partial\Omega,
# \end{align*}
# with $\Omega=[-\frac{1}{2},1]^2$ and choosing the forcing and the boundary conditions
# so that the exact solution is equal to
# \begin{align*}
#   u(x,y) &= xy\cos(\pi xy)
# \end{align*}

# First some setup code:
# <codecell>
try:
    get_ipython().magic(u'matplotlib nbagg')
    import matplotlib
    matplotlib.rc( 'image', cmap='rainbow' )
except:
    pass
import math
from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate
from dune.fem import parameter
from dune.vem import voronoiCells

from ufl import *
import dune.ufl

dune.fem.parameter.append({"fem.verboserank": 0})

# <markdowncell>
# we can compare different method, e.g., a lagrange space (on the # subtriangulation),
# a bounding box dg space and a conforming/non conforming VEM space

# <codecell>
methods = [ ### "[space,scheme,spaceKwrags]"
            ["lagrange","h1",{}],
            ["vem","vem",{"conforming":True}],
            ["vem","vem",{"conforming":False}],
            ["bbdg","bbdg",{}],
   ]
parameters = {"newton.linear.tolerance": 1e-12,
              "newton.linear.preconditioning.method": "ilu",
              "penalty": 40,  # for the bbdg scheme
              "newton.linear.verbose": True,
              "newton.verbose": True,
              }

# <markdowncell>
# Now we define the model starting with the exact solution:

# <codecell>
uflSpace = dune.ufl.Space(2, dimRange=1)
x = SpatialCoordinate(uflSpace)
exact = as_vector( [x[0]*x[1] * cos(pi*x[0]*x[1])] )

# next the bilinear form
u = TrialFunction(uflSpace)
v = TestFunction(uflSpace)
a = (inner(grad(u),grad(v))) * dx

# finally the right hand side and the boundary conditions
b = -div(grad(exact[0])) * v[0] * dx
dbc = [dune.ufl.DirichletBC(uflSpace, exact, i+1) for i in range(4)]


# <markdowncell>
# Now we define a grid build up of voronoi cells around $50$ random points

# <codecell>
constructor = cartesianDomain([-0.5,-0.5],[1,1],[1,1])
polyGrid = create.grid("agglomerate", voronoiCells(constructor,50) )

# <markdowncell>
# In general we can construct a `agglomerate` by providing a dictionary with
# the `vertices` and the `polygons`. The `voronoiCells` function creates
# such a dictonary using random seeds to generate voronoi cells which are
# cut off using the provided `cartesianDomain`. The seeds can be
# provided as list of points as second argument:
# ```
# voronoiCells(constructor, towers, fileName=None, load=False):
# ```
# If a `fileName` is provided the seeds will be written to disc and loaded
# from that file is `load=True`. As an example the output of
# `voronoiCells(constructor,5)` is
# ```
# {'polygons': [ [4, 5, 2, 3], [ 8, 10,  9,  7], [7, 9, 1, 3, 4],
#                [11, 10,  8,  0], [8, 0, 6, 5, 4, 7] ],
#  'vertices': [ [ 0.438, 1.  ],    [ 1. , -0.5 ],
#                [-0.5, -0.5  ],    [ 0.923, -0.5 ],
#                [ 0.248,  0.2214], [-0.5,  0.3027],
#                [-0.5,  1. ],      [ 0.407,0.4896],
#                [ 0.414,  0.525],  [ 1.,  0.57228],
#                [ 1., 0.88293],    [ 1.,  1. ] ] }
# ```
#
# Let's take a look at the grid with the 50 polygons triangulated

# <codecell>
@gridFunction(polyGrid, name="cells")
def polygons(en,x):
    return polyGrid.hierarchicalGrid.agglomerate(en)
polygons.plot()



# <markdowncell>
# We now define a function to compute the solution and the $L^2,H^1$ error
# given a grid and a space

# <codecell>
def compute(grid, space, schemeName):
    # solve the pde
    scheme = create.scheme(schemeName, [a==b, *dbc], space, solver="cg", parameters=parameters)
    df = space.interpolate([0],name="solution")
    info = scheme.solve(target=df)

    # compute the error
    edf = exact-df
    errors = [ math.sqrt(e) for e in
               integrate(grid, [inner(edf,edf),inner(grad(edf),grad(edf))], order=5) ]

    return df, errors

# <markdowncell>
# Finally we iterate over the requested methods and solve the problems

# <codecell>
for m in methods:
    space = create.space(m[0], polyGrid, order=2, dimRange=1, storage="istl", **m[2])
    dfs,errors = compute(polyGrid, space, m[1])
    print("Size: ",space.size, "L^2: ", errors[0], "H^1: ", errors[1])
    dfs.plot(gridLines=None)


print("END A")
# <markdowncell>
# ## Nonlinear elliptic problem
# We can easily set up a non linear problem
# \begin{align*}
# \end{align*}

# <codecell>
space = create.space("vem", polyGrid, order=2, dimRange=1, storage="istl",
                     conforming=True)
u = TrialFunction(space)
v = TestFunction(space)
x = SpatialCoordinate(space)
exact = as_vector ( [  (x[0] - x[0]*x[0] ) * (x[1] - x[1]*x[1] ) ] )
Dcoeff = lambda u: 1.0 + u[0]**2
a = (Dcoeff(u) * inner(grad(u), grad(v)) ) * dx
b = -div( Dcoeff(exact) * grad(exact[0]) ) * v[0] * dx
dbcs = [dune.ufl.DirichletBC(space, exact, i+1) for i in range(4)]
scheme = create.scheme("vem", [a==b, *dbcs], space, solver="cg", parameters=parameters)
solution = space.interpolate([0], name="solution")
info = scheme.solve(target=solution)
solution.plot(gridLines=None)
print("END B")
# <markdowncell>
# ## Linear Elasticity
# As final example we solve a linear elasticity equation usign a
# conforming VEM space:

# First we setup the domain
# <codecell>
L, W = 1, 0.2

constructor = cartesianDomain([0,0],[L,W],[1,1])
polyGrid = create.grid("agglomerate", voronoiCells(constructor,64) )
@gridFunction(polyGrid, name="cells")
def polygons(en,x):
    return polyGrid.hierarchicalGrid.agglomerate(en)
polygons.plot()

space = create.space("vem", polyGrid, order=2, dimRange=2, storage="istl",
                     conforming=True)

# <markdowncell>

# <codecell>
# some model constants
mu = 1
rho = 1
delta = W/L
gamma = 0.4*delta**2
beta = 1.25
lambda_ = beta
g = gamma

# clamped boundary on the left
x = SpatialCoordinate(space)
dbc = dune.ufl.DirichletBC(space, as_vector([0,0]), x[0]<1e-10)

# Define strain and stress
def epsilon(u):
    return 0.5*(nabla_grad(u) + nabla_grad(u).T)
def sigma(u):
    return lambda_*nabla_div(u)*Identity(2) + 2*mu*epsilon(u)

# Define the variational problem
u = TrialFunction(space)
v = TestFunction(space)
f = dune.ufl.Constant((0, -rho*g))
a = inner(sigma(u), epsilon(v))*dx
b = dot(as_vector([0,-rho*g]),v)*dx

# Compute solution
displacement = space.interpolate([0,0], name="displacement")
scheme = create.scheme("vem", [a==b, dbc], space, solver="cg", parameters=parameters)
info = scheme.solve(target=displacement)

# <markdowncell>
# Show the magnitude of the displacement field

# <codecell>
displacement.plot(gridLines=None)

# <markdowncell>
# We can easily plot the stress

# <codecell>
s = sigma(displacement) - (1./3)*tr(sigma(displacement))*Identity(2)  # deviatoric stress
von_Mises = sqrt(3./2*inner(s, s))
plot(von_Mises, grid=polyGrid, gridLines=None)

# <markdowncell>
# Finally we can plot the actual displaced beam

# <codecell>
from dune.fem.view import geometryGridView
position = space.interpolate( x+displacement, name="position" )
beam = geometryGridView( position )
beam.plot()



# <codecell>
try:
    get_ipython().magic(u'matplotlib inline # can also use notebook or nbagg')
except:
    pass
import math
from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate
from dune.fem import parameter
from dune.vem import voronoiCells

from ufl import *
import dune.ufl

dune.fem.parameter.append({"fem.verboserank": 0})

# <markdowncell>
# we can compare different method, e.g., a lagrange space (on the # subtriangulation),
# a bounding box dg space and a conforming/non conforming VEM space

# <codecell>
parameters = {"newton.linear.tolerance": 1e-12,
              "newton.linear.preconditioning.method": "ilu",
              "penalty": 40,  # for the bbdg scheme
              "newton.linear.verbose": False,
              }


# <markdowncell>
# As second example we solve a linear elasticity equation usign a
# conforming VEM space:

# First we setup the domain
# <codecell>
L, W = 1, 0.2

constructor = cartesianDomain([0,0],[L,W],[1,1])
polyGrid = create.grid("agglomerate", voronoiCells(constructor,64) )
@gridFunction(polyGrid, name="cells")
def polygons(en,x):
    return polyGrid.hierarchicalGrid.agglomerate(en)
polygons.plot()

space = create.space("vem", polyGrid, order=2, dimRange=2, storage="istl",
                     conforming=True)

# <markdowncell>

# <codecell>
# some model constants
mu = 1
rho = 1
delta = W/L
gamma = 0.4*delta**2
beta = 1.25
lambda_ = beta
g = gamma

# clamped boundary on the left
x = SpatialCoordinate(space)
dbc = dune.ufl.DirichletBC(space, as_vector([0,0]), x[0]<1e-10)

# Define strain and stress
def epsilon(u):
    return 0.5*(nabla_grad(u) + nabla_grad(u).T)
def sigma(u):
    return lambda_*nabla_div(u)*Identity(2) + 2*mu*epsilon(u)

# Define the variational problem
u = TrialFunction(space)
v = TestFunction(space)
f = dune.ufl.Constant((0, -rho*g))
a = inner(sigma(u), epsilon(v))*dx
L = dot(as_vector([0,-rho*g]),v)*dx

# Compute solution
displacement = space.interpolate([0,0], name="displacement")
scheme = create.scheme("vem", [a==L, dbc], space, solver="cg", parameters=parameters)
info = scheme.solve(target=displacement)

# <markdowncell>
# Show the magnitude of the displacement field

# <codecell>
displacement.plot(gridLines=None)

# <markdowncell>
# We can easily plot the stress

# <codecell>
s = sigma(displacement) - (1./3)*tr(sigma(displacement))*Identity(2)  # deviatoric stress
von_Mises = sqrt(3./2*inner(s, s))
plot(von_Mises, grid=polyGrid)

# <markdowncell>
# Finally we can plot the actual displaced beam

# <codecell>
from dune.fem.view import geometryGridView
position = space.interpolate( x+displacement, name="position" )
beam = geometryGridView( position )
beam.plot()



import os

path = os.path.join( os.path.dirname(__file__), "tutorial" )
execute  = "cp -r " + path + " "
execute += "vem_tutorial"
status = os.system(execute)
if status != 0: raise RuntimeError(status)

print("#################################################################")
print("## An example script is now located in the 'vem_tutorial' folder.")
print("#################################################################")



from __future__ import print_function

import math, sys, pickle

from ufl import *
import dune.ufl

from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem import parameter
from dune.fem.operator import linear
from dune.vem import voronoiCells

dimRange = 2
start    = 3
# polOrder, endEoc = 1,8
polOrder, endEoc = 2,8
# polOrder, endEoc = 3,6
# polOrder, endEoc = 4,5
# polOrder, endEoc = 5,4
# polOrder, endEoc = 6,4 # not working
methods = [ ### "[space,scheme,spaceKwrags]"
            ["lagrange","h1",{}],
            ["vem","vem",{"conforming":True}],
            ["vem","vem",{"conforming":False}],
            ["bbdg","bbdg",{}],
            # ["dgonb","dg",{}], # dg does not converge
   ]
parameters = {"newton.linear.tolerance": 1e-12,
              "newton.linear.verbose":True,
              "newton.tolerance": 1e-10,
              "newton.maxiterations": 10, # should finish in 1 for linear
              "newton.maxlinesearchiterations":50,
              "newton.verbose": False,
              "newton.linear.preconditioning.method": "amg-ilu",
              "penalty": 8*polOrder*polOrder
              }



dune.fem.parameter.append({"fem.verboserank": 0})

def plot(grid, solution):
    try:
        from matplotlib import pyplot
        from numpy import amin, amax, linspace

        triangulation = grid.triangulation(4)
        data = solution.pointData(4)

        levels = linspace(amin(data[:,0]), amax(data[:,0]), 256)

        pyplot.gca().set_aspect('equal')
        pyplot.triplot(grid.triangulation(), antialiased=True, linewidth=0.2, color='black')
        pyplot.tricontourf(triangulation, data[:,0], cmap=pyplot.cm.rainbow, levels=levels)
        pyplot.show()
    except ImportError:
        pass

def error(grid, df, interp, exact):
    edf = exact-df
    ein = exact-interp
    errors = create.function("ufl", grid, "error", 5,
            [ dot(edf,edf), dot(ein,ein),
              inner(grad(edf),grad(edf)), inner(grad(ein),grad(ein))
            ] ).integrate()
    return [ math.sqrt(e) for e in errors ]

h1errors  = []
l2errors  = []
spaceSize = []

def solve(grid,model,exact,space,scheme,spaceKwargs,order):
    print("SOLVING: ",space,scheme,spaceKwargs,flush=True)
    spc = create.space(space, grid, dimRange=dimRange, order=order,
            storage="istl", **spaceKwargs)
    name = space + "_".join(['']+[str(v) for v in spaceKwargs.values()])
    interpol = spc.interpolate( exact, "interpol_"+name )
    scheme = create.scheme(scheme, model, spc,
                solver="cg",
                parameters=parameters)
    df = spc.interpolate([0,]*dimRange,name=name)
    # info = {"linear_iterations":-1,"iterations":-1}
    info = scheme.solve(target=df)

    # import pickle,sys
    # linOp = linear(scheme)
    # scheme.jacobian(df,linOp)
    # pickle.dump([linOp.as_numpy,df.as_numpy], open( "dump"+str(dimRange), "wb" ) )
    # sys.exit(0)

    errors = error(grid,df,interpol,exact)
    print("Computed",name," size:",spc.size,
          "L^2 (s,i):", [errors[0],errors[1]],
          "H^1 (s,i):", [errors[2],errors[3]],
          "linear and Newton iterations:",
          info["linear_iterations"], info["iterations"],flush=True)
    global h1errors, l2errors, spaceSize
    l2errors  += [errors[0],errors[1]]
    h1errors  += [errors[2],errors[3]]
    spaceSize += [spc.size]
    return interpol, df

def compute(grid):
    uflSpace = dune.ufl.Space((grid.dimGrid, grid.dimWorld), dimRange, field="double")
    u = TrialFunction(uflSpace)
    v = TestFunction(uflSpace)
    x = SpatialCoordinate(uflSpace.cell())
    if True:
        ##### problem 1
        ### trivial Neuman bc
        exact  = as_vector([
                  cos(r*pi*x[0])*cos(r*pi*x[1]) for r in range(2,dimRange+2)
                ])
        #### zero boundary conditions
        exact *= x[0]*x[1]*(2-x[0])*(2-x[1])
        #### non zero and non trivial Neuman boundary conditions
        exact += as_vector([
                    sin(r*x[0]*r*x[1]) for r in range(1,dimRange+1)
                 ])
        H = lambda w: grad(grad(w))
        laplace = lambda w: H(w)[0,0] + H(w)[1,1]
        a = (inner(grad(u),grad(v))) * dx
        b = sum( [( -laplace(exact[r]) ) * v[r] * dx
                 for r in range(dimRange)] )
        # a += inner(u,v) * dx
        # b += inner(exact,v) * dx
    else:
        ##### problem 2: dummy quasilinear problem:
        exact = as_vector ( [  (x[0] - x[0]*x[0] ) * (x[1] - x[1]*x[1] ) ] )
        Dcoeff = lambda u: 1.0 + u[0]**2
        a = (Dcoeff(u)* inner(grad(u), grad(v)) ) * dx
        b = -div( Dcoeff(u) * grad(exact[0]) ) * v[0] * dx
    dbc = [dune.ufl.DirichletBC(uflSpace, exact, i+1) for i in range(4)]
    model = create.model("elliptic", grid, a==b, *dbc)
    dfs = []
    for m in methods:
        dfs += solve(grid,model,exact,*m,order=polOrder)

    @gridFunction(grid, name="cells")
    def polygons(en,x):
        return grid.hierarchicalGrid.agglomerate(en)
    grid.writeVTK(grid.hierarchicalGrid.agglomerate.suffix,
        pointdata=dfs, celldata=[polygons] )
    grid.writeVTK("s"+grid.hierarchicalGrid.agglomerate.suffix,subsampling=polOrder-1,
        pointdata=dfs, celldata=[polygons] )


for i in range(endEoc-start):
    print("*******************************************************")
    n = 2**(i+start)
    N = 2*n
    print("Test: ",n,N)
    constructor = cartesianDomain([0,0],[1,1],[N,N])
    # polyGrid = create.grid("agglomerate",constructor,[n,n])
    # polyGrid = create.grid("agglomerate",constructor,n*n)
    polyGrid = create.grid("agglomerate", voronoiCells(constructor,n*n,"voronoiseeds",True) )
    # polyGrid = create.grid("agglomerate",constructor)
    compute(polyGrid)
    if i>0:
        l = len(methods)
        for j in range(2*l):
            try:
                l2eoc = math.log( l2errors[2*l*i+j]/l2errors[2*l*(i-1)+j] ) / math.log(0.5)
            except ValueError:
                l2eoc = -1
            try:
                h1eoc = math.log( h1errors[2*l*i+j]/h1errors[2*l*(i-1)+j] ) / math.log(0.5)
            except ValueError:
                h1eoc = -1
            print("EOC",methods[int(j/2)][0],j,l2eoc,h1eoc)
    with open("errors_p"+str(polOrder)+".dump", 'wb') as f:
        pickle.dump([methods,spaceSize,l2errors,h1errors], f)
    print("*******************************************************")



import numpy
from dune.vem.qhull_2d import *
from dune.vem.min_bounding_rect import *

def rotatedBBox(points):
    points = numpy.array(points)
    if True: # rotated bbox
        # Find convex hull
        hull_points = qhull2D(points)
        # Reverse order of points, to match output from other qhull implementations
        hull_points = hull_points[::-1]
        # Find minimum area bounding rectangle
        (rot_angle, area, width, height, center_point, corner_points) = minBoundingRect(hull_points)
        e0 = corner_points[1]  - corner_points[0]
        e1 = corner_points[-1] - corner_points[0]
        # print( list(corner_points[0]), list(corner_points[2]), list(e0), list(e1), max(height,width), flush=True )
        return list(corner_points[0]), list(corner_points[2]), list(e0), list(e1), max(height,width)
    else: # axis aligned box
        xmin,xmax = min(points[:,0]), max(points[:,0])
        ymin,ymax = min(points[:,1]), max(points[:,1])
        return [xmin,ymin], [xmax,ymax], [xmax-xmin,0], [0,ymax-ymin], min(xmax-xmin,ymax-ymin)
